/* module for parsing robot configuration files */
/* Bruce Tanner -- Cerritos College */

/*
  find_config will look for a robot configuration file and return the
  rules that apply to this agent.

  Arguments:
  1 a hostname
  2 a port number
  3 a string containing the user agent name,
  4 a pointer to a result struct that will hold a linked list of
    disallowed directories

  Returns:
  -1 if no configuration file can be found
  0 if full access is allowed
  1 if some access is denied
  2 if all access is denied

*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#ifdef VMS
#include  <types.h>
#include  <unixio.h>
#include  <in.h>
#include  <socket.h>
#else
#include  <sys/socket.h>
#endif

#define NO_EXIST_CONFIG -1
#define FULL_ACCESS_CONFIG 0
#define PARTIAL_ACCESS_CONFIG 1
#define NO_ACCESS_CONFIG 2

extern int tcp_connect(char *, int);
extern int readline(int, char *);  /* external socket read routine */

typedef struct result_node {
    char *datafile;
    char *selector;
    struct result_node *next;
} result;

char *selector_prefix[] = {"", "0", "0/", "0/etc/"};

int find_file(char *, int, char *, char *, result *);
int find_name(int, char *, char *, char *, result *);
void parse_line(char *, char *, char *);
int match_index(char *, char *, char *);
char *match_datafile(char *, char *, char *);
int eval(int, result *);
int strcasecmp(char *, char *);

/* find a config file named 'myname'.ctl, veronica.ctl or robots.txt */
int find_config(char *host, int port, char *myname, result *results)
{

    char myfile[20];
    int ind, config;

    sprintf(myfile, "%s.ctl", myname);          /* create 'myname'.ctl */

    /* look for 'myname'.ctl */
    if ((config = find_file(host, port, myfile, myname,
                            results)) != NO_EXIST_CONFIG)
        return eval(config, results);           /* return results and config */

    /* try veronica.ctl */
    if ((config = find_file(host, port, "veronica.ctl", myname,
                            results)) != NO_EXIST_CONFIG)
        return eval(config, results);

    /* try robots.txt */
    if ((config = find_file(host, port, "robots.txt", myname,
                            results)) != NO_EXIST_CONFIG)
        return eval(config, results);

    return (NO_EXIST_CONFIG);                  /* no control block found */
}


/* find a file name prefixed with any of selector_prefix[] */
int find_file(char *host, int port, char *file, char *name, result *results)
{
    int ind, sock, config;

    for (ind = 0; ind < (sizeof(selector_prefix)/sizeof(char *)); ind++) {
        sock = tcp_connect(host, port);
        if ((config = find_name(sock, selector_prefix[ind], file, name,
                                results)) != NO_EXIST_CONFIG) {
            readline(sock, NULL);
            close(sock);
            return config;
        }
        readline(sock, NULL);
        close(sock);
    }
    return (NO_EXIST_CONFIG);                  /* no control block found */
}


/* search the prefix/file for name config block */
int find_name(int sock, char *prefix, char *file, char *name, result *results)
{
    char *df, selector[200], line[500], field[100], value[500];
    int status, config = NO_EXIST_CONFIG, our_block = 0;
    result *root;

    printf("Query %s%s\n", prefix, file);
    sprintf(selector, "0%s%s\n", prefix, file);  /* make selector */
    status = send(sock, selector, strlen(selector), 0);
    root = results;                        /* save pointer to root result */

    while (status = readline(sock, line)) {
        if (line[0] == '.')
            break;
        /* are we at end of our block? */
        if (((line[0] == '\n') || (line[0] == '\r')) && our_block)
            return config;                     /* we're done */

        parse_line(line, field, value);        /* no, parse line */
        our_block |= ((strcasecmp(field, "user-agent") == 0) &&
                      ((strcasecmp(value, name) == 0) || (strcmp(value, "*") == 0)));
        if (our_block) {
            if (config == NO_EXIST_CONFIG)  /* make transition from  no exist */
                config = FULL_ACCESS_CONFIG;  /* to full access */
            if ((status = match_index(field, value, name)) == NO_ACCESS_CONFIG)
                return (NO_ACCESS_CONFIG);    /* 'name'index: no */
            if (df = match_datafile(field, value, name))
                root->datafile = df;
            if ((strcasecmp(field, "disallow") == 0) && strlen(value)) {
                config = PARTIAL_ACCESS_CONFIG; /* remember 'some disallowed' */
                results->next = (result *)calloc(1, sizeof(result));
                results->selector = (char *)calloc(strlen(value) + 1, sizeof(char));
                strcpy(results->selector, value);
                results = results->next;
            }
        }
    }
    return config;
}


/* normalize a string */
char *normal(char *str)
{
    char *cp;

    /* remove leading spaces */
    while (*str && (*str <= ' '))
        strcpy(str, str + 1);

    /* remove trailing spaces, crlf */
    while (*str && (str[strlen(str) - 1] <= ' '))
        str[strlen(str) - 1] = '\0';

    return (str);
}


/* parse a line of text into field name and value */
void parse_line(char *line, char *field, char *value)
{
    char *cp;

    if ((*line == '#') ||                      /* comment or */
        ((cp = strchr(line, ':')) == NULL)) {  /* no colon */
        *field = '\0';                         /* clear field, value */
        *value = '\0';
        return;
    }

    *cp = '\0';                     /* mark end of field name */
    strcpy(field, normal(line));    /* copy field name */
    strcpy(value, normal(cp + 1));  /* copy value */
}


/* is this a 'name'index field? */
int match_index(char *field, char *value, char *name)
{
    char str[50];

    strcpy(str, name);              /* make the field name */
    strcat(str, "index");
    if (strcasecmp(str, field))     /* is this a 'name'index field? */
        return (NO_EXIST_CONFIG);   /* no */      
    if (strcasecmp(value, "no") == 0)   /* index: no ? */
        return (NO_ACCESS_CONFIG);  /* all access denied */
    return (FULL_ACCESS_CONFIG);    /* allow full access */
}


char *match_datafile(char *field, char *value, char *name)
{
    char *cp, str[50];

    strcpy(str, name);              /* make the field name */
    strcat(str, "-datafile");
    if (strcasecmp(str, field))     /* is this a 'name'-datafile field? */
        return (NULL);              /* no */
    cp = (char *)calloc(strlen(value) + 1, sizeof(char));
    strcpy(cp, value);
    return (cp);
}


/* see if we need to convert 'some disallowed' (2) to 'all disallowed' (0) */
int eval(int config, result *results)
{
    result *ptr;
    int ind;
    static char *all_selector[] = {"1/", "/", "1[000000]", "[000000]"};
 
    if (config != PARTIAL_ACCESS_CONFIG)           /* not partial? */
        return config;                             /* no, return that config */

    for (ptr = results; ptr->selector; ptr = ptr->next)
        for (ind = 0; ind < (sizeof(all_selector)/sizeof(char *)); ind++)
            if (strcmp(ptr->selector, all_selector[ind]) == 0)  /* all? */
                return (NO_ACCESS_CONFIG);              /* yes, disallow all */
    return (config);
}


/* case-independent compare */
int strcasecmp(char *s1, char *s2)
{
    int ind;

    for (ind = 0; s1[ind] && s2[ind]; ind++)
        if (_tolower(s1[ind]) != _tolower(s2[ind]))
            break;
    return (_tolower(s1[ind]) - _tolower(s2[ind]));
}
